#include "offset.h"
#include "kjv.h"
#include "web.h"
#include "dr.h"
#include <ctype.h>
#include <errno.h>
#include <regex.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

typedef struct { int book, chapter, verse; } ref;
typedef struct { ref start, end; } range;

#define VALUE(x) (((x)[0] * 0x10000 + (x)[1] * 0x100 + (x)[2]) & ~0x800000)
#define FLAGS(x) ((((x)[0] * 0x10000) & 0x800000) >> 23)
#define ZERO(r) ((r)->book = (r)->chapter = (r)->verse = 0)
#define SIZE(o) ((o)->length)
#define START(o) ((o)->offset)

int bb_flag = 0;
int canon_flag = 0;
int case_flag = 0;
int cgi_flag = 0;
int context_flag = 0;
int escape_flag = 0;
int extended_flag = 0;
int html_flag = 0;
int kjv_flag = 0;
int line_flag = 0;
int link_flag = 0;
int noformat_flag = 0;
int pretty_flag = 0;
int quiet_flag = 0;
int indent_flag = 0;
int quote_flag = 0;
int simple_flag = 0;
int verse_flag = 0;
int version_flag = 0;
int web_flag = 0;
int dr_flag = 0;
int width = 79;
int bb_save = 0;
int canon_save = 0;
int escape_save = 0;
char *scriptname = "/cgi-bin/bible";

char *searchpat = 0;

unsigned char output_buffer[1024];
size_t output_pos = 0;

bible *b = NULL;
/* should be 1 more than number of Bibles that can be selected at once */
bible *bbls[] = { NULL, NULL, NULL, NULL };
int bblcount = 0;


struct flag {
	int c;
	char *name;
	int *flag;
} flags[] = {
	/* "bCcDeEFhiKlnpQqs:vw:W" */
	{ 'b', "BB", &bb_flag },
	{ 'c', "CONTEXT", &context_flag },
	{ 'D', "DOUAY", &dr_flag },
	{ 'e', "ESCAPE", &escape_flag },
	{ 'E', "EXTENDED", &extended_flag },
	{ 'F', "SIMPLE", &simple_flag },
	{ 'h', "HTML", &html_flag },
	{ 'i', "IGNORE", &case_flag },
	{ 'I', "INDENT", &indent_flag },
	{ 'K', "KJV", &kjv_flag },
	{ 'l', "LINE", &line_flag },
	{ 'L', "LINK", &link_flag },
	{ 'n', "NOFORMAT", &noformat_flag },
	{ 'N', "CANON", &canon_flag },
	{ 'p', "PRETTY", &pretty_flag },
	{ 'Q', "QUOTE", &quote_flag },
	{ 'q', "QUIET", &quiet_flag },
	{ 'v', "VERSE", &verse_flag },
	{ 'V', "VERSION", &version_flag },
	{ 'W', "WEB", &web_flag },
	{ '\0', 0, 0 },
};

char *
optionname(int c) {
	struct flag *f = flags;
	for (f = flags; f->c; ++f)
		if (f->c == c)
			return f->name;
	return "INVALID";
}

int *
optionflag(int c) {
	static int dummy;
	struct flag *f = flags;
	dummy = 0;
	for (f = flags; f->c; ++f)
		if (f->c == c)
			return f->flag;
	return &dummy;
}

void
error(char *fmt, ...) {
	va_list ap;
	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
	fputc('\n', stderr);
	if (cgi_flag) {
		puts("<H3>Error: ");
		va_start(ap, fmt);
		vfprintf(stdout, fmt, ap);
		va_end(ap);
		puts("</H3>");
	}
}

int
refcmp(ref *a, ref *b) {
	if (!a || !b)
		return 0;
	if (a->book < b->book)
		return -1;
	if (a->book > b->book)
		return 1;
	if (a->chapter < b->chapter)
		return -1;
	if (a->chapter > b->chapter)
		return 1;
	if (a->verse < b->verse)
		return -1;
	if (a->verse > b->verse)
		return 1;
	return 0;
}

int
value(unsigned char x[3]) {
	return VALUE(x);
}

int
ref_start(ref r) {
	coffset *oc;
	voffset *ov;
	oc = &b->oc[START(&b->ob[r.book])];
	ov = &b->ov[START(&oc[r.chapter])];
	return(VALUE(ov[r.verse].offset));
}

int
ref_flags(ref r) {
	coffset *oc;
	voffset *ov;
	oc = &b->oc[START(&b->ob[r.book])];
	ov = &b->ov[START(&oc[r.chapter])];
	return(FLAGS(ov[r.verse].offset));
}

range
ref2range(ref r) {
	range r2;
	r2.start = r;
	r2.end = r;
	return r2;
}

char *
range2url(range r) {
	static char url[2048];
	char *s = url;
	struct flag *f;
	*s = '\0';
	strncat(s, scriptname, 1024);
	strcat(s, "?");
	s += strlen(s);
	for (f = flags; f->c; ++f) {
		if (*(f->flag)) {
			strcat(s, f->name);
			strcat(s, "=on&amp;");
			s += strlen(s);
		}
	}
	sprintf(s, "RANGE=%s%%20%d%%3A%d",
		b->book_names[r.start.book],
		r.start.chapter,
		r.start.verse);
	s += strlen(s);
	if (r.start.book != r.end.book) {
		sprintf(s, "-%s%%20%d%%3A%d",
			b->book_names[r.end.book],
			r.end.chapter,
			r.end.verse);
	} else if (r.start.chapter != r.end.chapter) {
		sprintf(s, "-%d%%3A%d",
			r.end.chapter,
			r.end.verse);
	} else if (r.start.verse != r.end.verse) {
		sprintf(s, "-%d",
			r.end.verse);
	}
	return url;
}

int
matchbook(unsigned char *s, int len) {
	int i;
	int lastmatch = 0;
	int match = 0;
	unsigned char *t, *u;

	if (!len)
		return -1;
	for (i = 0; b->book_names[i]; ++i) {
		u = s;
		t = (unsigned char *) b->book_names[i];
		if (tolower(*u) != tolower(*t))
			continue;
		match = 1;
		while (*u && *t && (u < s + len)) {
			while (isspace(*u))
				++u;
			while (isspace(*t))
				++t;
			if (!*t) {
				match = 0;
				break;
			}
			if (!*u || (u >= s + len))
				break;
			if (tolower(*u) == tolower(*t)) {
				if (isdigit(*u)) {
					/* get rid of st/nd/rd */
					if ((*u == '1' &&
					     !strncmp((char *)u+1, "st", 2)) ||
					    (*u == '2' &&
					     !strncmp((char *)u+1, "nd", 2)) ||
					    (*u == '3' &&
					     !strncmp((char *)u+1, "rd", 2)))
						u+=2;
				}
				++u;
				++t;
				continue;
			} else {
				match = 0;
				break;
			}
		}
		if (match) {
			if (lastmatch) {
				error("Ambiguous reference: '%s' could match either '%s' or '%s'.",
						s,
						b->book_names[lastmatch],
						b->book_names[i]);
				return -1;
			}
			lastmatch = i;
		}
	}
	return lastmatch;
}

int
convert_book(int b, bible *b1, bible *b2) {
	char *name;
	int i;
	if (!b1 || !b2)
		return -1;
	if (b < 1 || b > SIZE(b1->ob))
		return -1;
	name = b1->book_names[b];
	for (i = 0; i <= SIZE(b2->ob); ++i) {
		if (!strcmp(name, b2->book_names[i]))
			return i;
	}
	return -1;
}

void
usage(void) {
	/* "bCcDeEFhiKlnpQqs:vVw:W" */
	/* This insane ugliness is necessary because the string exceeds the
	 * 509-character string limit, but it has to be all one thing for the
	 * HTML mode, in case anyone figures out a way to get a usage message
	 * in HTML mode. */
	error("%s\n%s\n%s",
		"usage: bible [-bCcDeEFhiKlNnpQqvW] [-w w] [-s re] book chapter:verse [- b c:v]",
	      "\t-b\tUse [] markup for rendering\n"
	      "\t-C\tEmulate CGI mode\n"
	      "\t-c\tDisplay 5-verse context\n"
	      "\t-D\tUse Douay-Rheims version\n"
	      "\t-e\tEscape HTML entities\n"
	      "\t-h\tGenerate HTML output\n"
	      "\t-i\tCase-insensitive searching\n"
	      "\t-I\tIndent output\n"
	      "\t-K\tUse King James Version\n"
	      "\t-l\tPrint each verse on a new line",
	      "\t-N\tCanonical formatting\n"
	      "\t-n\tNo formatting\n"
	      "\t-p\tPretty-print\n"
	      "\t-Q\tFormat as quote\n"
	      "\t-q\tSuppress headers\n"
	      "\t-s\tSearch regexp\t-E extended\t-F plaintext\n"
	      "\t-v\tNumber verses"
	      "\t-V\tPrint version/copyright info"
	      "\t-W\tUse World English Bible\n"
	      "\t-w\tSpecify width\n"
	      "Specifying both multiple Bibles prints all.  Specifying none uses KJV."
	      );
}

void
normalize(ref *r) {
	boffset *ob;
	coffset *oc;
	if (!r)
		return;
	if (r->book < 1) {
		/* The hubris of attempting to peer past "In the
		 * beginning" goes unpunished. */
		r->book = 1;
		r->chapter = 1;
		r->verse = 1;
		return;
	}
	if (r->book > SIZE(b->ob)) {
		r->book = SIZE(b->ob);
		r->chapter = SIZE(&b->ob[r->book]);
		r->verse = SIZE(&b->oc[b->ob[r->book].offset + r->chapter]);
		return;
	}
	ob = &b->ob[r->book];
	/* book is in range to start with */
	while (r->chapter < 0) {
		--r->book;
		if (r->book < 1) {
			/* The hubris of attempting to peer past "In the
			 * beginning" goes unpunished. */
			r->book = 1;
			r->chapter = 1;
			r->verse = 1;
			return;
		}
		--ob;
		r->chapter += SIZE(ob);
	}
	while (r->chapter > SIZE(ob)) {
		r->chapter -= SIZE(ob);
		++r->book;
		while (SIZE(&b->oc[b->ob[r->book].offset + r->chapter]) == 0)
			++r->chapter;
		if (r->book > SIZE(b->ob)) {
			r->book = SIZE(b->ob);
			r->chapter = SIZE(&b->ob[r->book]);
			r->verse = SIZE(&b->oc[b->ob[r->book].offset + r->chapter]);
			return;
		}
		++ob;
	}
	oc = &b->oc[START(ob) + r->chapter];
	/* chapter is in range to start with */
	while (r->verse < 1) {
		--r->chapter;
		--oc;
		if (r->chapter < 1) {
			--r->book;
			--ob;
			if (r->book < 1) {
				/* The hubris of attempting to peer past "In the
				 * beginning" goes unpunished. */
				r->book = 1;
				r->chapter = 1;
				r->verse = 1;
				return;
			}
			r->chapter += SIZE(ob);
		}
		r->verse += SIZE(oc);
	}
	while (r->verse > SIZE(oc)) {
		r->verse -= SIZE(oc);
		++r->chapter;
		++oc;
		if (r->chapter > SIZE(ob)) {
			r->chapter = 1;
			r->book++;
			++ob;
			if (r->book > SIZE(b->ob)) {
				r->book = SIZE(b->ob);
				r->chapter = SIZE(&b->ob[r->book]);
				r->verse = SIZE(&b->oc[b->ob[r->book].offset + r->chapter]);
				return;
			}
		}
	}
}

range
getrange(char *refname) {
	range r = { { 0 } };
	char *ch, *v, *end;
	int book;

	if (!refname)
		return r;
	end = strchr(refname, '-');
	if (end) {
		*end++ = 0;
	}

	while (isspace((unsigned char) *refname))
		++refname;
	/* chapter is first number not in initial position */
	ch = strpbrk(refname+1, "0123456789");
	v = strrchr(refname, ':');
	book = matchbook((unsigned char *) refname,
		ch ? (ch - refname) : strlen(refname));
	if (book == -1) {
		return r;
	}
	r.start.book = book;
	r.start.chapter = ch ? atoi(ch) : 1;
	if (v) {
		r.start.verse = atoi(v+1);
	} else {
		r.start.verse = 1;
	}

	r.end.book = r.start.book;
	r.end.chapter = ch ? r.start.chapter : SIZE(&b->ob[r.start.book]);
	r.end.verse = r.start.verse;

	if (!v) {
		r.end.chapter++;
		r.end.verse = 0;
	}

	if (end) {
		while (isspace((unsigned char) *end))
			++end;
		/* now we have to guess what this is. */
		if (strspn(end, "1234567890-: ") == strlen(end)) {
			if ((v = strchr(end, ':')) != 0) {
				r.end.chapter = atoi(end);
				r.end.verse = atoi(v+1);
			} else {
				if (r.end.verse < 1) {
					r.end.chapter = atoi(end)+1;
				} else {
					r.end.verse = atoi(end);
				}
			}
		} else {
			ch = strpbrk(end+1, "0123456789");
			v = strrchr(end, ':');
			book = matchbook((unsigned char *) end,
				ch ? (ch - end) : strlen(end));
			if (book == -1) {
				return r;
			}
			r.end.book = book;
			if (ch) {
				r.end.chapter = atoi(ch);
			} else {
				r.end.chapter = 0;
				r.end.book++;
				r.end.verse = 0;
			}
			if (v) {
				r.end.verse = atoi(v+1);
			} else {
				r.end.chapter++;
				r.end.verse = 0;
			}
		}
	}

	normalize(&r.start);
	normalize(&r.end);
	return r;
}

void
flush(void) {
	fwrite(output_buffer, 1, output_pos, stdout);
	output_pos = 0;
}

void
csend(int c) {
	int lastpos = output_pos;
	if (canon_flag) {
		if (isspace(c) || ispunct(c)) {
			return;
		}
		if (isalpha(c))
			c = toupper(c);
	}
	/* another layer of HTMLization for HTML output */
	if (escape_flag) {
		switch (c) {
		case '<':
			output_buffer[output_pos++] = '&';
			output_buffer[output_pos++] = 'l';
			output_buffer[output_pos++] = 't';
			output_buffer[output_pos++] = ';';
			break;
		case '>':
			output_buffer[output_pos++] = '&';
			output_buffer[output_pos++] = 'g';
			output_buffer[output_pos++] = 't';
			output_buffer[output_pos++] = ';';
			break;
		case '&':
			output_buffer[output_pos++] = '&';
			output_buffer[output_pos++] = 'a';
			output_buffer[output_pos++] = 'm';
			output_buffer[output_pos++] = 'p';
			output_buffer[output_pos++] = ';';
			break;
		case '\n':
			if (cgi_flag) {
				output_buffer[output_pos++] = '<';
				output_buffer[output_pos++] = 'b';
				output_buffer[output_pos++] = 'r';
				output_buffer[output_pos++] = '>';
			}
			output_buffer[output_pos++] = c;
			break;
		default:
			output_buffer[output_pos++] = c;
		}
	} else {
		if ((cgi_flag || html_flag) && c == '\n') {
			output_buffer[output_pos++] = '<';
			output_buffer[output_pos++] = 'b';
			output_buffer[output_pos++] = 'r';
			output_buffer[output_pos++] = '>';
		}
		output_buffer[output_pos++] = c;
	}
	if (c == '\n') {
		flush();
	}
	if (output_pos > width) {
		unsigned char *s;
		output_buffer[output_pos] = 0;
		s = (unsigned char *) strrchr((char *) output_buffer, ' ');
		if (s) {
			fwrite(output_buffer, 1, s - output_buffer, stdout);
			if (canon_flag || !cgi_flag) {
				putchar('\n');
				/* skip the space */
				++s;
			}
			memmove(output_buffer, s, output_pos - (s - output_buffer));
			output_pos -= s - output_buffer;
		} else {
			fwrite(output_buffer, 1, lastpos, stdout);
			if (canon_flag || !cgi_flag) {
				putchar('\n');
			}
			memmove(output_buffer, output_buffer + lastpos, output_pos - 78);
			output_pos -= lastpos;
		}
	}
}

void
send(char *s) {
	while (*s)
		csend(*s++);
}

void
sendf(char *fmt, ...) {
	char buffer[2048];
	va_list ap;
	va_start(ap, fmt);
	vsnprintf(buffer, 2048, fmt, ap);
	va_end(ap);
	send(buffer);
}

void
ital(void) {
	if (noformat_flag)
		return;
	canon_flag = 0;
	if (bb_flag) {
		send("[i]");
	}
	if (html_flag) {
		send("<em>");
	}
	if (!html_flag && !bb_flag) {
		send("[");
	}
	canon_flag = canon_save;
}

void
unital(void) {
	if (noformat_flag)
		return;
	canon_flag = 0;
	if (html_flag) {
		send("</em>");
	}
	if (bb_flag) {
		send("[/i]");
	}
	if (!html_flag && !bb_flag) {
		send("]");
	}
	canon_flag = canon_save;
}

void
bold(void) {
	if (noformat_flag)
		return;
	canon_flag = 0;
	if (bb_flag) {
		send("[b]");
	}
	if (html_flag) {
		send("<strong>");
	}
	canon_flag = canon_save;
}

void
unbold(void) {
	if (noformat_flag)
		return;
	canon_flag = 0;
	if (html_flag) {
		send("</strong>");
	}
	if (bb_flag) {
		send("[/b]");
	}
	canon_flag = canon_save;
}

void
newline(void) {
	if (canon_flag)
		return;
	if (html_flag && escape_flag) {
		send("<br>\n");
	} else {
		csend('\n');
	}
}

size_t
expand(char *dest, unsigned char *src, size_t len) {
	char *t;
	for (t = dest; *src && (t - dest < len); ++src) {
		if (b->abbrevs[*src & 0x7f]) {
			t += expand(t, b->abbrevs[*src & 0x7f], len - (t - dest));
		} else {
			switch (*src & 0x7f) {
			case '[':
			case ']':
				/* do nothing */
				break;
			default:
				*t++ = *src & 0x7f;
			}
		}
		if (*src & 0x80)
			*t++ = ' ';
	}
	return t - dest;
}

void
dwrite(unsigned char *s) {
	for (; *s; ++s) {
		if (b->abbrevs[*s & 0x7f]) {
			dwrite(b->abbrevs[*s & 0x7f]);
			if (*s & 0x80)
				csend(' ');
		} else {
			switch (*s & 0x7f) {
			case '\n':
				fprintf(stderr, "actually saw \\n in dwrite\n");
				newline();
				break;
			case '[':
				ital();
				break;
			case ']':
				unital();
				break;
			default:
				csend(*s & 0x7f);
				break;
			}
			if (*s & 0x80)
				csend(' ');
		}
	}
}

void
printbook(int book) {
	if (book < 1 || book > SIZE(b->ob))
		sendf("Unknown book %d", book);
	if (pretty_flag) {
		dwrite((unsigned char *) b->book_titles[book]);
	} else {
		send(b->book_names[book]);
	}
}

int
expandverse(ref r, char *dest, size_t len) {
	return expand(dest, b->text + ref_start(r), len);
}

void
printverse(ref r) {
	dwrite(b->text + ref_start(r));
}

void
printchaprange(range r) {
	int printed_links = 0;
	if (r.start.book != r.end.book || r.start.chapter != r.end.chapter) {
		fprintf(stderr, "%d:%d, %d:%d\n",
			r.start.book, r.start.chapter, r.end.book,
			r.end.chapter);
		error("I thought this would all be in the same chapter, but it wasn't!");
		return;
	}
	newline();
	if (cgi_flag && !context_flag) {
		canon_flag = 0;
		escape_flag = 0;
		context_flag = 1;
		sendf("<a href=\"%s\">(Context)</a> ", range2url(r));
		context_flag = 0;
		canon_flag = canon_save;
		escape_flag = escape_save;
		printed_links = 1;
	}
	if (cgi_flag && bblcount == 1) {
		canon_flag = 0;
		escape_flag = 0;
		sendf("<a href=\"%s", range2url(r));
		if (!kjv_flag) {
			sendf("&amp;KJV=on");
		}
		if (!web_flag) {
			sendf("&amp;WEB=on");
		}
		if (!dr_flag) {
			sendf("&amp;DOUAY=on");
		}
		sendf("\">(All versions)</a> ");
		printed_links = 1;
	}
	if (printed_links)
		newline();
	canon_flag = 0;
	escape_flag = escape_save;
	if (quote_flag) {
		if (bb_flag) {
			send("[quote");
		} else if (html_flag) {
			send("<blockquote>");
		} else {
			send("   ");
		}
	} else if (indent_flag) {
		if (bb_flag) {
			send("[indent]");
		} else if (html_flag) {
			send("<blockquote>");
		} else {
			send("   ");
		}
	}
	if (!quiet_flag) {
		canon_flag = 0;
		if (bb_flag && quote_flag) {
			csend('=');
		} else {
			bold();
			canon_flag = 0;
		}
		if (cgi_flag) {
			escape_flag = 0;
			sendf("<a href=\"%s\">", range2url(r));
			escape_flag = escape_save;
		}
		if (link_flag) {
			if (bb_flag)
				sendf("[url=%s]", range2url(r));
			else if (html_flag)
				sendf("<a href=\"%s\">", range2url(r));
		}
		canon_flag = canon_save;
		printbook(r.start.book);
		if (!canon_flag)
			send(pretty_flag ? ", Chapter " : " ");
		canon_flag = 0;
		sendf("%d", r.start.chapter);
		canon_flag = canon_save;
		if (!canon_flag) {
			if (pretty_flag)
				sendf(", Verse%s ", (r.start.verse != r.end.verse) ? "s" : "");
			else
				csend(':');
		}
		canon_flag = 0;
		sendf("%d", r.start.verse);
		if (r.start.verse != r.end.verse) {
			sendf("-%d", r.end.verse);
		}
		if (cgi_flag) {
			escape_flag = 0;
			send("</a>");
			escape_flag = escape_save;

		}
		canon_flag = canon_save;
		if (link_flag) {
			if (bb_flag)
				sendf("[/url]");
			else if (html_flag)
				sendf("</a>");
		}
		if (!(bb_flag && quote_flag))
			unbold();
	}
	if (quote_flag && bb_flag) {
		/* close the quote tag */
		send("]");
	}
	newline();
	canon_flag = canon_save;
	while (refcmp(&r.start, &r.end) <= 0) {
		if (ref_flags(r.start)) {
			/* this double-spaces for paragraphs, but I can
			 * live with that.
			 */
			if (!line_flag)
				newline();
			canon_flag = 0;
			if (pretty_flag) {
				if (html_flag || (cgi_flag && !escape_flag)) {
					send("&nbsp;&nbsp;&nbsp;");
				} else if (cgi_flag && escape_flag) {
					escape_flag = 0;
					send("&nbsp;&nbsp;&nbsp;");
					escape_flag = 1;
				} else {
					send("   ");
				}
			}
			canon_flag = canon_save;
		}
		if (verse_flag) {
			canon_flag = 0;
			if (cgi_flag) {
				escape_flag = 0;
				sendf("<a href=\"%s\">", range2url(ref2range(r.start)));
				escape_flag = escape_save;
			}
			if (pretty_flag)
				bold();
			ital();
			sendf("%d", r.start.verse);
			unital();
			if (pretty_flag)
				unbold();
			if (cgi_flag) {
				escape_flag = 0;
				canon_flag = 0;
				send("</a>");
				escape_flag = escape_save;
			}
			canon_flag = canon_save;
			csend(' ');

		}
		printverse(r.start);
		if (refcmp(&r.start, &r.end) < 0) {
			++r.start.verse;
			normalize(&r.start);
		} else {
			break;
		}
		if (line_flag)
			newline();
		else
			csend(' ');
	}
	if (quote_flag) {
		canon_flag = 0;
		if (bb_flag)
			send("[/quote]");
		else if (html_flag)
			send("</blockquote>\n");
		canon_flag = canon_save;
	} else if (indent_flag) {
		canon_flag = 0;
		if (bb_flag)
			send("[/indent]");
		else if (html_flag)
			send("</blockquote>\n");
		canon_flag = canon_save;
	}
	canon_flag = 0;
	newline();
	canon_flag = canon_save;
	flush();
}

void
printonerange(range r) {
	int context_save = context_flag;

	normalize(&r.start);
	if (r.start.book == 0) {
		error("Couldn't find any such verse.");
		return;
	}
	if (context_flag) {
		r.start.verse -= 5;
		r.end.verse += 5;
		context_flag = 0;
		normalize(&r.start);
	}
	normalize(&r.end);
	if (refcmp(&r.start, &r.end) > 0) {
		ref tmp = r.start;
		r.start = r.end;
		r.end = tmp;
	}
	if (r.start.book != r.end.book || r.start.chapter != r.end.chapter) {
		range chap;
		chap.start.book = r.start.book;
		chap.start.chapter = r.start.chapter;
		chap.start.verse = r.start.verse;
		chap.end = r.start;
		/* normalize will take care of this */
		++chap.end.chapter;
		chap.end.verse = 0;
		do {
			normalize(&chap.start);
			normalize(&chap.end);
			/* shouldn't happen.  Hah! */
			if (refcmp(&chap.start, &chap.end) > 0)
				return;
			printchaprange(chap);
			chap.start.verse = 1;
			++chap.start.chapter;
			/* this moves us to end of next chapter, which is
			 * just before beginning of chapter after next
			 */
			chap.end.chapter += 2;
			chap.end.verse = 0;
			normalize(&chap.start);
			normalize(&chap.end);
			if (chap.end.book == r.end.book && chap.end.chapter == r.end.chapter) {
				chap.end.verse = r.end.verse;
			}
		} while(refcmp(&chap.end, &r.end) < 0);
		/* and finally the last one */
		printonerange(chap);
		context_flag = context_save;
		return;
		/* we never actually do the real work */
	} else {
		printchaprange(r);
	}
	context_flag = context_save;
}

void
printrange(range r) {
	int newstart, newend;
	int bbl;
	int save_kjv = kjv_flag;
	int save_web = web_flag;
	int save_dr = dr_flag;
	bible *save_b = b;

	canon_flag = 0;

	for (bbl = 0; bbl < bblcount; ++bbl) {
		newstart = convert_book(r.start.book, b, bbls[bbl]);
		newend = convert_book(r.end.book, b, bbls[bbl]);
		b = bbls[bbl];
		kjv_flag = (b == &kjv);
		web_flag = (b == &web);
		dr_flag = (b == &dr);
		bb_flag = 0;
		bold();
		if (newstart == -1 || newend == -1) {
			error("Missing book in %s", b->bible_name);
			unbold();
			bb_flag = bb_save;
		} else {
			range r2 = r;
			send(b->bible_name);
			unbold();
			bb_flag = bb_save;
			canon_flag = canon_save;
			r2.start.book = newstart;
			r2.end.book = newend;
			printonerange(r2);
		}
		b = save_b;
	}
	kjv_flag = save_kjv;
	web_flag = save_web;
	dr_flag = save_dr;
}

void
searchrange(char *range_text) {
	range r;
	range searchme;
	ref expandme;
	regex_t pat;
	int eflags = REG_NOSUB;
	int result;
	/* not sure this is big enough, but we'll try it */
	char searchbuf[4096];
	char *sbp, *second;
	int matches = 0;
	int bbl;

	if (searchpat == 0) {
		error("Can't search without a search pattern.");
		return;
	}
	if (extended_flag)
		eflags |= REG_EXTENDED;
	else if (simple_flag)
		eflags |= REG_NOSPEC;
	if (case_flag)
		eflags |= REG_ICASE;

	if ((result = regcomp(&pat, searchpat, eflags)) != 0) {
		error("regcomp failed, search pattern invalid [%d]", result);
		return;
	}
	pat.re_endp = 0;

	for (bbl = 0; bbl < bblcount; ++bbl) {
		char *t = range_text ? strdup(range_text) : 0;
		b = bbls[bbl];
		/* escape this in case search pattern contains HTML */
		escape_save = escape_flag;
		escape_flag = 1;
		sendf("Searching %s (%s) for '%s':\n", b->bible_name,
			(t && strlen(t)) ? t : "whole Bible", searchpat);
		escape_flag = escape_save;
		r = getrange(t);
		free(t);

		if (r.start.book == 0) {
			r.start.book = 1;
			r.start.chapter = 1;
			r.start.verse = 1;
			r.end.book = SIZE(b->ob);
			r.end.chapter = SIZE(&b->ob[r.end.book]);
			r.end.verse = SIZE(&b->oc[START(&b->ob[r.end.book]) + r.end.chapter]);
		}

		searchme.start = r.start;

	found:
		searchme.end = searchme.start;

		do {
			expandme = searchme.start;
			sbp = searchbuf;
			sbp += expandverse(expandme, sbp, 4096 - (sbp - searchbuf));
			*sbp = '\0';
			/* check single verse */
			if (!regexec(&pat, searchbuf, 0, NULL, 0)) {
				++matches;
				printonerange(searchme);
				/* move ahead another verse to avoid reprinting match */
				++searchme.start.verse;
				normalize(&searchme.start);
				/* and now restart the loop from just after
				 * the verse we found, unless we're at the
				 * end
				 */
				if (refcmp(&expandme, &searchme.start))
					goto found;
			}

			do {
				++expandme.verse;
				normalize(&expandme);
				searchme.end = expandme;
				/* if it's the same verse (e.g., rev 22:21), we're done */
				if (refcmp(&expandme, &searchme.start) == 0)
					break;

				*sbp++ = ' ';
				second = sbp;
				sbp += expandverse(expandme, sbp, 4096 - (sbp - searchbuf));
				*sbp = '\0';
				/* check second verse */
				if (!regexec(&pat, second, 0, NULL, 0)) {
					++matches;
					searchme.start = searchme.end;
					printonerange(searchme);
					/* move ahead another verse to avoid reprinting match */
					++searchme.start.verse;
					normalize(&searchme.start);
					/* ... so we matched, so we don't need to check
					 * with the combined verses */
					if (refcmp(&expandme, &searchme.start))
						goto found;
					else
						break;
				}

				if (!regexec(&pat, searchbuf, 0, NULL, 0)) {
					++matches;
					printonerange(searchme);
					/* move ahead another verse to avoid reprinting match */
					++searchme.end.verse;
					normalize(&searchme.end);
					searchme.start = searchme.end;
					/* ... so we matched, so we don't need to check
					 * with the combined verses */
					goto found;
				}

				/* now move second verse to front of buffer: */
				memmove(searchbuf, second, (sbp - second));
				sbp = searchbuf + (sbp - second);
				searchme.start = searchme.end;
				++searchme.end.verse;
				normalize(&searchme.end);
				/* if we didn't find anything, bump the first verse and
				 * try again */
			} while (refcmp(&searchme.end, &r.end) <= 0);
		} while (refcmp(&searchme.start, &r.end) < 0);
	}
	regfree(&pat);
	if (matches) {
		sendf("\nTotal matches: %d\n", matches);
		flush();
	} else {
		error("No matches, sorry!");
	}
}

void
cb(int option, char *descr) {
	char *name = optionname(option);
	int checked = *optionflag(option);
	printf("<TD><INPUT TYPE=\"CHECKBOX\" NAME=\"%s\"%s></TD><TH ALIGN=\"LEFT\">%s</TH>",
		name, checked ? " CHECKED" : "", descr);
}

int
cgimain(void) {
	char *request = strdup(getenv("QUERY_STRING"));
	char *s, *t, *u;
	char *name, *value;
	char *lastrange = "";

	if (getenv("SCRIPT_NAME"))
		scriptname = strdup(getenv("SCRIPT_NAME"));

	cgi_flag = 1;

	puts("Content-Type: text/html\n\n");

	puts("<HTML><HEAD><TITLE>HTML Bible</TITLE></HEAD><BODY>");
	if (strlen(request)) {
		range r = { { 0 }, { 0 } };
		char *tmp;
		puts("<H3>Here ya go:</H3>");
		s = request;
		while ((t = strsep(&s, "&")) != 0) {
			name = t;
			strsep(&t, "=");
			value = t;
			u = t;
			while (*t) {
				if (*t == '%') {
					unsigned int c = 0;
					sscanf(t + 1, "%2x", &c);
					*u = c;
					t += 2;
				} else if (*t == '+') {
					*u = ' ';
				} else {
					*u = *t;
				}
				++t, ++u;
			}
			*u = '\0';
			if (!strcmp(name, "RANGE")) {
				lastrange = strdup(value);
			} else if (!strcmp(name, "SEARCH")) {
				if (strlen(value))
					searchpat = strdup(value);
				else
					searchpat = 0;
			} else {
				struct flag *f;
				for (f = flags; f->c; ++f) {
					if (!strcmp(name, f->name)) {
						switch (f->c) {
						case 'K':
							if (!kjv_flag) {
								kjv_flag = 1;
								bbls[bblcount++] = &kjv;
							}
							break;
						case 'D':
							if (!dr_flag) {
								dr_flag = 1;
								bbls[bblcount++] = &dr;
							}
							break;
						case 'W':
							if (!web_flag) {
								web_flag = 1;
								bbls[bblcount++] = &web;
							}
							break;
						default:
							*(f->flag) = 1;
						}
					}
				}
			}
		}
		canon_save = canon_flag;
		escape_save = escape_flag;
		bb_save = bb_flag;
		if (bblcount == 0) {
			kjv_flag = 1;
			bbls[bblcount++] = &kjv;
		}
		bbls[bblcount] = NULL;
		b = bbls[0];
		tmp = strdup(lastrange);
		if (searchpat) {
			searchrange(tmp);
		} else {
			r = getrange(tmp);
			printrange(r);
		}
		free(tmp);
	} else {
		/* how about some sensible defaults? */
		pretty_flag = 1;
		html_flag = 1;
	}
	puts("<H3>Whaddya need?</H3>");
	printf("<FORM METHOD=\"GET\" ACTION=\"%s\">\n", scriptname);
	puts("<TABLE BORDER=\"0\"><TR>");
	puts("<TH COLSPAN=\"2\"><FONT SIZE=\"+1\">Verse range</FONT></TH>");
	puts("<TH COLSPAN=\"2\"><FONT SIZE=\"+1\">Search text</FONT></TH></TR><TR>");
	puts("</TR><TR>");
	printf("<TD COLSPAN=\"2\" ALIGN=\"CENTER\"><INPUT TYPE=\"TEXT\" NAME=\"RANGE\" SIZE=\"30\" VALUE=\"%s\"></TD>\n",
		lastrange);
	printf("<TD COLSPAN=\"2\" ALIGN=\"CENTER\"><INPUT TYPE=\"TEXT\" NAME=\"SEARCH\" SIZE=\"30\" VALUE=\"%s\"></TD>\n",
		searchpat ? searchpat : "");
	puts("</TR><TR>");
	puts("<TH COLSPAN=\"2\" ALIGN=\"CENTER\">e.g., <i>Romans 8:38-39</i></TH>");
	puts("<TH COLSPAN=\"2\" ALIGN=\"CENTER\">e.g., <i>nor things to come</i></TH></TR><TR>");
	puts("</TR><TR>");
	puts("<TH COLSPAN=\"4\"><FONT SIZE=\"+2\">Options:</FONT></TH></TR><TR>");
	puts("<TH COLSPAN=\"4\" ALIGN=\"LEFT\"><FONT SIZE=\"+1\">Display Options:</FONT></TH></TR><TR>");
	cb('q', "Suppress headers");
	cb('p', "Pretty-print output");
	puts("</TR><TR>");
	cb('v', "Print verse numbers");
	cb('l', "New line for each verse");
	puts("</TR><TR>");
	puts("<TH COLSPAN=\"4\" ALIGN=\"LEFT\"><FONT SIZE=\"+1\">Formatting Options:</FONT></TH></TR><TR>");
	cb('b', "Format as vBcode");
	cb('h', "Format as HTML");
	puts("</TR><TR>");
	cb('Q', "Format as quote");
	cb('e', "Escape HTML entities");
	puts("</TR><TR>");
	cb('I', "Format as indent");
	cb('L', "Produce hyperlinks");
	puts("</TR><TR>");
	cb('n', "Suppress formatting");
	cb('N', "Canonical formatting");
	puts("</TR><TR>");
	puts("<TH COLSPAN=\"4\" ALIGN=\"LEFT\"><FONT SIZE=\"+1\">Search Options:</FONT></TH></TR><TR>");
	cb('F', "Simple search (no regex)");
	cb('E', "Extended regexes");
	puts("</TR><TR>");
	cb('i', "Ignore case (much slower)");
	cb('c', "Print surrounding context");
	puts("</TR><TR>");
	puts("<TH COLSPAN=\"4\" ALIGN=\"LEFT\"><FONT SIZE=\"+1\">Bible Options:</FONT> (default is KJV)</TH></TR><TR>");
	cb('K', "King James Version");
	cb('W', "World English Bible");
	puts("</TR><TR>");
	cb('D', "Douay-Rheims Bible");
	puts("</TR></TABLE>");
	puts("<INPUT TYPE=\"SUBMIT\" NAME=\"GO\" VALUE=\"Go to it!\">");
	puts("</FORM>");
	puts("<P><i>Bible CGI program written by Peter Seebach, and placed in the public domain.  Bibles incorporated from text believed to be in the public domain, found on the internet.  Source available by request.  No rights reserved, and no warranty, explicit or implicit, is given.</i></P>");
	puts("</BODY></HTML>");
	return 0;
}

int
main(int argc, char **argv) {
	int o;
	range r;
	char buffer[256] = "";
	struct flag *f;

	if (getenv("REQUEST_METHOD")) {
		return cgimain();
	}
	while ((o = getopt(argc, argv, "bCcDeEFhiIKlLNnpQqs:vVw:W")) != -1) {
		switch (o) {
		default:
			for (f = flags; f->c; ++f) {
				if (o == f->c) {
					*(f->flag) = 1;
					break;
				}
			}
			if (f->c)
				break;
			/* else FALLTHROUGH */
		case '?':
			usage();
			exit(EXIT_FAILURE);
			break;
		case 'b':
			bb_flag = 1;
			break;
		case 'c':
			context_flag = 1;
			break;
		case 'C':
			/* just pretend */
			cgi_flag = 1;
			break;
		case 'D':
			if (!dr_flag) {
				dr_flag = 1;
				bbls[bblcount++] = &dr;
			}
			break;
		case 'e':
			escape_flag = 1;
			break;
		case 'E':
			extended_flag = 1;
			break;
		case 'F':
			simple_flag = 1;
			break;
		case 'h':
			html_flag = 1;
			break;
		case 'i':
			case_flag = 1;
			break;
		case 'I':
			indent_flag = 1;
			break;
		case 'K':
			if (!kjv_flag) {
				kjv_flag = 1;
				bbls[bblcount++] = &kjv;
			}
			break;
		case 'l':
			line_flag = 1;
			break;
		case 'L':
			link_flag = 1;
			break;
		case 'n':
			noformat_flag = 1;
			break;
		case 'p':
			pretty_flag = 1;
			break;
		case 'Q':
			quote_flag = 1;
			break;
		case 'q':
			quiet_flag = 1;
			break;
		case 's':
			searchpat = strdup(optarg);
			break;
		case 'v':
			verse_flag = 1;
			break;
		case 'w':
			width = atoi(optarg);
			if (width < 1)
				width = 79;
			/* see size of output_buffer above */
			if (width > 1000)
				width = 1000;
			break;
		case 'W':
			if (!web_flag) {
				web_flag = 1;
				bbls[bblcount++] = &web;
			}
			break;
		}
	}
	canon_save = canon_flag;
	escape_save = escape_flag;
	bb_save = bb_flag;
	argv += optind;
	argc -= optind;

	/* okay, now to try to read stuff */
	if (argc < 1 && !searchpat) {
		usage();
		exit(EXIT_FAILURE);
	}

	while (argv[0] && strlen(buffer) < 256) {
		strncat(buffer, argv[0], 255 - strlen(buffer));
		++argv;
		--argc;
	}
	if (bblcount == 0) {
		kjv_flag = 1;
		bbls[bblcount++] = &kjv;
	}
	bbls[bblcount] = NULL;
	b = bbls[0];
	if (searchpat)
		searchrange(buffer);
	else {
		r = getrange(buffer);
		printrange(r);
	}

	return 0;
}

