#!/usr/pkg/bin/perl -w
use IO::Handle;
use integer;
$swirlpos = 0;
$swirl = "|/-\\";
my @books = ();
my @titles = ();
my %words = ();
$prefix = $ARGV[0];
open(BOOKS, "$prefix/books.dat");
while (<BOOKS>) {
	chomp;
	s/{.*}{//;
	s/}$//;
	push(@books, $_);
}
close BOOKS;
$byte = 0;
$data = "";
$oldchapter = 0;
$oldbook = 0;
$verse = 0;
$maxbook = 99;
foreach $i (0..127) {
	$chars[$i] = 0;
}
print "Reading raw file to build histograms...\n";
STDOUT->autoflush(1);

$book = 0;
while (<STDIN>) {
	chop;
	s/\} */\}/;
	push(@rdata, $_);
	if (/^{[A-Z]/) {
		++$book;
		last if $book > $maxbook;
		s/^{//;
		s/}$//;
		printf "\rReading: %-69s", $_;
		foreach $c (split(//)) {
			++$chars[ord($c)];
		}
		next;
	} else {
		s/^{.*}//;
	}
	foreach $c (split(//)) {
		++$chars[ord($c)];
	}
	foreach $w (m/\b\S{2}\B/g) {
		++$words{$w};
	}
	foreach $w (m/\B\S{2}\b/g) {
		++$words{$w};
	}
	foreach $w (m/\b\S{2}\b/g) {
		++$words{$w};
	}
	foreach $w (m/\b\S{3}\b/g) {
		++$words{$w};
	}
	foreach $w (m/\b\S{3}\B/g) {
		++$words{$w};
	}
	foreach $w (m/\B\S{3}\b/g) {
		++$words{$w};
	}
	foreach $w (m/\b\S{4}\b/g) {
		++$words{$w};
	}
	foreach $w (m/\b\S{4}\B/g) {
		++$words{$w};
	}
	foreach $w (m/\B\S{4}\b/g) {
		++$words{$w};
	}
	foreach $w (m/\b\S{5}\b/g) {
		++$words{$w};
	}
	foreach $w (m/\b\S{5}\B/g) {
		++$words{$w};
	}
	foreach $w (m/\B\S{5}\b/g) {
		++$words{$w};
	}
	foreach $w (m/\S{6,}/g) {
		++$words{$w};
	}
}
printf "\rReading: %-69s\n", "done.";

open(HIST, ">hist$prefix.dat");
$count = 0;
foreach $c (0..127) {
	if ($chars[$c]) {
		print HIST sprintf("%02x (%s): %d\n", $c, chr($c), $chars[$c]);
		++$count;
	}
}
print HIST "(Total $count chars used)\n";
close HIST;
print "$count characters in use, finding abbreviations...\n";

# The way we count multiple hits could in some cases result in a combination
# being found more often than it occurs.
foreach $w (sort { length $a <=> length $b } keys %words) {
	next if length($w) < 2;
	last if length($w) > 4;
	if (length($w) == 2) {
		$words{$w} = $words{$w} / 2;
	}
	if (length($w) == 3) {
		$words{$w} = $words{$w} * 2 / 3;
	}
	if (length($w) == 4) {
		$words{$w} = $words{$w} * 3 / 4;
	}
}
foreach $w (sort { length $b <=> length $a } keys %words) {
	if ($words{$w} < 200) {
		$l = length $w;
		foreach $i (1..length $w) {
			$s = substr($w, 0, $l - $i);
			if (defined $words{$s}) {
				$words{$s} += $words{$w};
				last;
			}
		}
		delete $words{$w};
	}
}

$count = 0;
foreach $w (sort { length $b <=> length $a } keys %words) {
	next if !defined $words{$w};
	last if length $w < 3;
	++$count;
	if ($count % 10 == 0) {
		printf("\r%-30.30s %-15s", $w, $words{$w} . swirl() . "\b");
	}
	foreach $y (sort { length $b <=> length $a } keys %words) {
		next unless defined $y;
		last if (length $y) <= length($w);
		if ($y =~ /\Q$w/) {
			# if $w is 6 characters or longer, it matches only
			# when it's the whole word, so it didn't get credit
			# for superstrings it would also match
			if (length $w >= 6) {
				$words{$w} += $words{$y};
			}
			if ($words{$y} > ($words{$w} * (length($w) - 1) / (length($y) - 1))) {
				# reduce scoring of w to reflect that it's
				# a substring of y, which we like better
				$words{$w} -= $words{$y};
				if ($words{$w} < 0) {
					delete $words{$w};
					last;
				}
			} else {
				# reduce scoring of y to reflect that picking 
				# w will save you some of y's length.
				$words{$y} -= $words{$y} * ((length $w) - 1) / ((length $y) - 1);
				if ($words{$y} < 0) {
					delete $words{$y};
				}
			}
		}
	}
}
print "done.\n";
open(WORDS, ">word$prefix.dat");
$count = 0;
foreach $w (keys %words) {
	print WORDS "$words{$w} $w\n" if $words{$w} > 20;
}
close WORDS;

my @abbrevs = sort { ((length($b)-1)*$words{$b}) <=> ((length($a)-1)*$words{$a}) } keys %words;

open(ABBR, ">abbr$prefix.dat");
$saved = 0;
foreach $c (1..127) {
	next if chr($c) eq "\n";
	if ($chars[$c] == 0) {
		my $w = shift @abbrevs;
		# ran out!
		last unless defined $w;
		$t = $words{$w} * (length($w) - 1);
		$saved += $t;
		print ABBR "$c -> '$w' (x$words{$w}, -$t)\n";
		$abbrevs{$w} = chr($c);
		$spaceabbrevs{$w} = chr($c + 128);
		$rabbrevs{$c} = $w;
	}
}
close ABBR;

@abby = sort { length $b <=> length $a } keys %abbrevs;

print "Found ", scalar @abby, " abbreviations, saving $saved bytes.\n";

$book = 0;
$uncompressed = 0;
$tchapter = 0;
$tverse = 0;
$oldbook = 0;
$oldverse = 0;
$oldchapter = 0;
foreach $_ (@rdata) {
	if (/^{[A-Z]/) {
		s/^{//;
		s/}$//;
		$length{"$oldbook"} = $oldchapter;
		$length{"$oldbook:$oldchapter"} = $oldverse;
		$titles[$book] = $_;
		$bookname = $books[$book];
		++$book;
		$tchapter += $oldchapter;
		$offset{$book} = $tchapter;
		last if ($book > $maxbook);
		$oldchapter = 0;
		$oldbook = $book;
		printf("\rProcessing... %-50s", "$bookname ($book)");
		# and update
	} else {
		$tag = $_;
		$text = $_;
		$text =~ s/{.*}//;
		$tag =~ s/{//;
		$tag =~ s/}.*$//;
		$flags = "";
		($chapter, $verse, $flags) = split(':', $tag);
		if (defined $flags && $flags eq "p") {
			$breaks{"$book:$chapter:$verse"} = 1;
		}
		if ($chapter > $oldchapter) {
			print swirl(), "\b";
			if ($oldchapter < $chapter-1) {
				foreach $i ($oldchapter..$chapter-1) {
					$offset{"$book:$i"} = $tverse - 1;
					$length{"$book:$i"} ||= 0;
				}
			}
			$tverse += $oldverse;
			$offset{"$book:$chapter"} = $tverse;
			$length{"$book:$oldchapter"} = $oldverse;
			$oldchapter = $chapter;
		}
		$offset{"$book:$chapter:$verse"} = $byte;
		$uncompressed += length $text;
		foreach $w (@abby) {
			$text =~ s/\Q$w /$spaceabbrevs{$w}/g;
			$text =~ s/\Q$w/$abbrevs{$w}/g;
		}
		while ($text =~ m/([\001-\177] )/g) {
			$c = chr(ord($&) + 128);
			$text =~ s/\Q$&/$c/g;
		}
		$text .= "\0";
		$data .= $text;
		$byte += length $text;
		$oldverse = $verse;
	}
}
$compressed = $byte;
printf("\rProcessing... %-52s\n", "done.");
printf("%d bytes text compressed to %d bytes (%d%%)\n",
	$uncompressed, $compressed, ($compressed * 100 / $uncompressed));
$length{"$oldbook"} = $oldchapter;
$length{"$oldbook:$oldchapter"} = $oldverse;

print "Storing header... ";
open(CHEADER, ">$prefix.h");
print CHEADER "extern boffset ob_$prefix\[\];\n";
print CHEADER "extern coffset oc_$prefix\[\];\n";
print CHEADER "extern voffset ov_$prefix\[\];\n";
print CHEADER "extern char * book_names_$prefix\[\];\n";
print CHEADER "extern char * book_titles_$prefix\[\];\n";
print CHEADER "extern unsigned char text_$prefix\[\];\n";
print CHEADER "extern unsigned char *abbrevs_$prefix\[\];\n";
print CHEADER "bible $prefix = { ob_$prefix, oc_$prefix, ov_$prefix, text_$prefix, abbrevs_$prefix, book_names_$prefix, book_titles_$prefix, \"$ARGV[1]\" };\n";
close CHEADER;
print "done.\n";
open(CHEADER, ">$prefix.c");
print CHEADER "#include \"offset.h\"\n";

$book_header = "boffset ob_$prefix\[\] = {\n";
$book_header .= "\t{ 0, $oldbook },\n";
$chapter_header = "coffset oc_$prefix\[\] = {\n";
$chapter_header .= "\t{ 0, 0 },\n";

print CHEADER "voffset ov_$prefix\[\] = {\n";
print CHEADER "\t{ { 0 } },\n";

foreach $b (1..$oldbook) {
	last unless defined $offset{$b};
	$s = $offset{"$b"};
	$l = $length{"$b"};
	$book_header .=  "\t{ $s, $l },\n";
	foreach $c (1..$length{"$b"}) {
		$s = $offset{"$b:$c"};
		$l = $length{"$b:$c"};
		if (!defined $offset{"$b:$c"}) {
			print STDERR "No offset recorded for $b:$c\n";
			$s = 65535;
		}
		if (!defined $length{"$b:$c"}) {
			print STDERR "No length recorded for $b:$c\n";
			$l = 65535;
		}
		$chapter_header .= "\t{ $s, $l },\n";
		foreach $v (1..$length{"$b:$c"}) {
			$s = $offset{"$b:$c:$v"};
			if (!defined $s) {
				print STDERR "No offset recorded for $b:$c:$v\n";
			}
			if ($breaks{"$b:$c:$v"}) {
				$s += 0x800000;
			}
			$s = threebyte($s);
			print CHEADER "\t{ $s },\n";
		}
	}
}
print CHEADER "\t{ { 0 } }\n};\n\n";
print CHEADER "\n";

print CHEADER $chapter_header, "\t{ 0 }\n};\n\n";
print CHEADER $book_header, "\t{ 0 }\n};\n\n";
print CHEADER "char *book_names_$prefix\[\] = {\n\t\"NO_BOOK_0\",\n";
foreach $b (@books) {
	print CHEADER "\t\"$b\",\n";
}
print CHEADER "\t0\n};\n\n";

print CHEADER "char *book_titles_$prefix\[\] = {\n\t\"NO_BOOK_0\",\n";
foreach $b (@titles) {
	print CHEADER "\t\"$b\",\n";
}
print CHEADER "\t0\n};\n\n";

print CHEADER "\nunsigned char *abbrevs_$prefix\[\] = {\n\t0,\n";
foreach $c (1..127) {
	if ($rabbrevs{$c}) {
		$t = $rabbrevs{$c};
		$t =~ s/"/\\"/g;
		print CHEADER "\t(unsigned char *) \"$t\",\n";
	} else {
		print CHEADER "\t0,\n";
	}
}
print CHEADER "};\n\n";

print "Storing text...   \b\b\b";
print CHEADER "unsigned char text_$prefix\[$compressed\] = {\n\t";
$count = 0;
while ($count < length $data) {
	print CHEADER ord(substr($data, $count++, 1)), ", ";
	if (($count % 16) == 0) {
		print CHEADER "\n\t";
	}
	if (($count % 4096) == 0) {
		print swirl(), "\b";
	}
}
print CHEADER "\n};\n\n";
print "$compressed bytes done.  \n";

close CHEADER;

sub
swirl {
	$x = substr($swirl, $swirlpos++, 1);
	$swirlpos %= 4;
	return $x;
}

sub
threebyte {
	my $value = shift;
	return "{ 0, 0, 0 }" unless defined $value;
	$hi = $value / 65536;
	$med = ($value % 65536) / 256;
	$lo = $value % 256;
	return "{ $hi, $med, $lo }";
}
